from __future__ import annotations

import importlib
import json
import logging
import math
import os
from uuid import UUID

from gridappsd import GridAPPSD
from rdflib import Graph, URIRef

import cimgraph.queries.sparql as sparql
from cimgraph.databases import ConnectionInterface, QueryResponse

_log = logging.getLogger(__name__)


class GridappsdConnection(ConnectionInterface):

    def __init__(self, connection_params):
        self.cim_profile = connection_params.cim_profile
        self.cim = importlib.import_module('cimgraph.data_profile.' + self.cim_profile)
        self.namespace = connection_params.namespace
        self.iec61970_301 = connection_params.iec61970_301
        self.connection_params = connection_params

        if connection_params.host:
            os.environ['GRIDAPPSD_ADDRESS'] = connection_params.host
        else:
            os.environ['GRIDAPPSD_ADDRESS'] = 'localhost'

        if connection_params.port:
            os.environ['GRIDAPPSD_PORT'] = connection_params.port
        else:
            os.environ['GRIDAPPSD_PORT'] = '61613'

        if connection_params.database:
            self.database = connection_params.database
        else:
            self.database = 'powergridmodel'

        if not connection_params.url:
            self.connection_params.url = 'http://localhost:8889/bigdata/namespace/kb/sparql'

        os.environ['GRIDAPPSD_APPLICATION_ID'] = 'cimantic-graphs'
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STARTED'
        os.environ['GRIDAPPSD_USER'] = 'app_user'
        os.environ['GRIDAPPSD_PASSWORD'] = '1234App'

        self.gapps = None

        try:
            self.rdfs_profile = Graph(store='Oxigraph')
            path = os.path.dirname(self.cim.__file__)
            self.rdfs_profile.parse(f'{path}/{self.cim_profile}.rdfs', format='xml')
            self.reverse = URIRef('http://iec.ch/TC57/1999/rdf-schema-extensions-19990926#inverseRoleName')
        except:
            self.rdfs_profile = None

    def connect(self):
        if not self.gapps:
            self.gapps = GridAPPSD()

    def disconnect(self):
        self.gapps.disconnect()

    def execute(self, query_message: str) -> QueryResponse:
        self.connect()
        response = self.gapps.query_data(query_message, database_type=self.database, timeout=30)
        return response['data']

    def create_new_graph(self, container: object) -> dict[type, dict[UUID, object]]:
        """
        Create a new graph structure for a CIM EquipmentContainer object.
        The method uses a SPARQL query to obtain all terminals in the graph,
        along with all nodes and conducting equipment associated with each
        terminal. This forms the baseline knowledge graph for the GraphModel.
        If a graph is specified, the new objects will be added to the existing
        graph. Otherwise, a new graph will be created from scratch.

        Args:
            container (object): The container object for which the graph is created.
            graph (dict, optional): Graph of CIM objects, grouped by class and UUID.

        Returns:
            dict: Graph consisting of types and UUID mapped to object instances.
        """
        graph = {}
        self.add_to_graph(graph=graph, obj=container)
        # Get all nodes, terminal, and equipment by
        sparql_message = sparql.get_all_nodes_from_container(container, self.connection_params)
        query_output = self.execute(sparql_message)
        graph = self.parse_node_query(graph, query_output)
        return graph

    def create_distributed_graph(self, area: object, graph: dict = {}) -> Graph:
        self.add_to_graph(graph=graph, obj=area)

        if not isinstance(area, self.cim.SubSchedulingArea):
            _log.error(f'Area object is not a SubSchedulingArea')
            raise TypeError('Area object is not a SubSchedulingArea')

        sparql_message = sparql.get_all_nodes_from_area(area, self.connection_params)
        # Execute SPARQL query
        query_output = self.execute(sparql_message)
        # Parse query results and create new graph
        graph = self.parse_node_query(graph, query_output)
        return graph

    def parse_node_query(self, graph: dict, query_output: dict) -> Graph:
        """
        Parse the results of a node query to update a graph structure.

        Args:
            graph (dict): Graph to be updated.
            query_output (dict): Query output to be parsed.

        Returns:
            dict: Updated graph structure.
        """
        # Iterate through all rows of query output
        for result in query_output['results']['bindings']:

            # Associated conducting equipment are JSON-LD strings
            eq = json.loads(result['Equipment']['value'])
            eq_id = eq['@id']
            eq_class = eq['@type']

            # If equipment class is in data profile, add it to the graph also
            if eq_class in self.cim.__all__:
                eq_class = eval(f'self.cim.{eq_class}')
                equipment = self.create_object(graph, eq_class, eq_id)
            else:
                # If it is not in the profile, log it as a missing class
                _log.warning(
                    f'object class missing from data profile: {eq_class}')
                continue
            if 'ConnectivityNode' in result:
                # Get uri strings of nodes and terminals
                node_mrid = result['ConnectivityNode']['value']
                term_mrid = result['Terminal']['value']
                # Add each object to graph
                node = self.create_object(graph, self.cim.ConnectivityNode, node_mrid)
                terminal = self.create_object(graph, self.cim.Terminal, term_mrid)
                # Associate the node and equipment with the terminal
                if terminal not in equipment.Terminals:
                    equipment.Terminals.append(terminal)
                if terminal not in node.Terminals:
                    node.Terminals.append(terminal)
                # Associate the terminal with the equipment and node
                setattr(terminal, 'ConnectivityNode', node)
                setattr(terminal, 'ConductingEquipment', equipment)
            if 'Measurement' in result:
                # Associated conducting equipment are JSON-LD strings
                meas = json.loads(result['Measurement']['value'])
                meas_id = meas['@id']
                meas_class = meas['@type']
                # Get uri strings of nodes and terminals
                if meas_class in self.cim.__all__:
                    meas_class = eval(f'self.cim.{meas_class}')
                    measurement = self.create_object(graph, meas_class, meas_id)

        return graph

    def build_graph_from_list(
            self, graph,
            mrid_list: list[str]) -> dict[type, dict[str, object]]:
        for index in range(math.ceil(len(mrid_list) / 100)):
            eq_mrids = mrid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct loaders>sparql python script based on class name
            sparql_message = sparql.get_all_nodes_from_list(
                eq_mrids, self.namespace)
            # print(sparql_message)
            query_output = self.execute(sparql_message)
            graph = self.parse_node_query(graph, query_output)
        return graph

    def get_edges_query(self, graph: dict[type, dict[UUID, object]],
                        cim_class: type) -> str:

        eq_mrids = list(graph[cim_class].keys())[0:100]
        sparql_message = sparql.get_all_edges_sparql(graph, cim_class, eq_mrids,
                                                     self.connection_params)

        return sparql_message

    def get_all_edges(self, graph: dict[type, dict[UUID, object]], cim_class: type) -> None:
        uuid_list = list(graph[cim_class].keys())
        for index in range(math.ceil(len(uuid_list) / 100)):
            eq_mrids = uuid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct queries>sparql python script based on class name
            sparql_message = sparql.get_all_edges_sparql(graph,
                cim_class, eq_mrids, self.connection_params)
            #execute sparql query
            query_output = self.execute(sparql_message)
            self.edge_query_parser(query_output, graph, cim_class)

    def get_all_attributes(self, graph: dict[type, dict[str, object]],
                           cim_class: type):
        mrid_list = list(graph[cim_class].keys())
        num_nodes = len(mrid_list)
        for index in range(math.ceil(len(mrid_list) / 100)):
            eq_mrids = mrid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct loaders>sparql python script based on class name
            sparql_message = sparql.get_all_attributes_sparql(
                cim_class, eq_mrids, self.connection_params)
            #execute sparql query
            query_output = self.execute(sparql_message)
            self.edge_query_parser(query_output, graph, cim_class)

    def edge_query_parser(self, query_output: QueryResponse,
                          graph: Graph, cim_class: type, expand_graph=True) -> list[object]:
        """
        Parse the results of an edge query to update a graph structure.

        Args:
            query_output (QueryResponse): Query output to be parsed.
            graph (dict): Graph structure.
            cim_class (type): The CIM class for which to parse edges.
            expand_graph (bool, optional): Whether to expand the graph with new edges. Defaults to True.
        """
        new_edges = []
        for result in query_output['results']['bindings']:

            if result['attribute']['value'] != 'type':  #skip 'type' and other single attributes

                uri = result['identifier']['value']  #get mRID
                identifier = UUID(uri.strip('_').lower())
                attribute = result['attribute']['value']  #edge attribute
                value = result['value']['value']  #get edge value
                if 'edge' in result:  #check if association
                    edge = json.loads(result['edge']['value'])
                    edge_mRID = edge['@id']
                    edge_class = edge['@type']
                    if edge_class in self.cim.__all__:
                        edge_class = eval(f'self.cim.{edge_class}')
                    else:
                        _log.warning(f'Class {edge_class} not in data profile')
                        continue

                    if expand_graph:
                        edge_object = self.create_edge(graph, cim_class, identifier,
                                         attribute, edge_class, edge_mRID)
                        new_edges.append(edge_object)
                    else:
                        self.create_value(graph, cim_class, identifier, attribute, value)

                elif self.namespace in value:  #check if enumeration
                    enum_text = value.split(self.namespace)[1]
                    enum_text = enum_text.split('>')[0]
                    enum_class = enum_text.split('.')[0]
                    enum_value = enum_text.split('.')[1]

                    if enum_class in self.cim.__all__:  # if enumeration
                        edge_enum = eval(f'self.cim.{enum_class}(enum_value)')
                        new_edges.append(edge_enum)
                        association = self.check_attribute(
                            cim_class, attribute)
                        if association is not None:
                            setattr(graph[cim_class][identifier], association, edge_enum)
                else:
                    association = self.check_attribute(cim_class, attribute)
                    if association is not None:
                        new_edges.append(value)
                        self.create_value(graph, cim_class, identifier, attribute, value)
        return new_edges


    def upload(self, graph: dict[type, dict[str, object]]) -> None:
        for cim_class in graph.keys():
            for obj in graph[cim_class].values():
                query = sparql.upload_triples_sparql(obj, self.connection_params)
                self.execute(query)

    def get_from_triple(self, subject:object, predicate:str, graph: Graph = {}) -> list[object]:
        if not graph:
            self.add_to_graph(subject, graph)
        # Generate SPARQL query for user-specified triple string
        sparql_message = sparql.get_triple_sparql(subject, predicate, self.connection_params)
        # Execute SPARQL query
        query_output = self.execute(sparql_message)
        # Parse the query output
        new_edges = self.edge_query_parser(query_output, graph, subject.__class__)
        return new_edges

    def get_object(self, mrid: str, graph: dict = {}) -> object:
        """
        Retrieve an object from the Blazegraph database using its mRID.

        Args:
            mrid (str): The mRID of the object to be retrieved.
            graph (dict, optional): The graph database to store the fetched object. Defaults to {}.

        Returns:
            object: The retrieved object.
        """
        # Use sparql module to build get correct query string
        sparql_message = sparql.get_object_sparql(mrid, self.connection_params)
        # Execute query
        query_output = self.execute(sparql_message)
        obj = None
        # Parse query results to get the correct CIM object
        for result in query_output['results']['bindings']:
            uri = result['identifier']['value']  # uri / mRID string
            obj_class = result['obj_class']['value']  # class type
            # If equipment class is in data profile, create a new object
            if obj_class in self.cim.__all__:
                class_type = eval(f'self.cim.{obj_class}')  # get type
                obj = self.create_object(graph, class_type, uri)  # get object
            else:
                # If it is not in the profile, log it as a missing class
                _log.warning(
                    f'object class missing from data profile: {obj_class}')
                continue

        return obj



from __future__ import annotations

from uuid import UUID

from cimgraph.databases import ConnectionParameters


def get_triple_sparql(obj:object, attribute:str, connection_params: ConnectionParameters) -> str:
    """
    Generates SPARQL query string to find predicate for RDF triple
    Args:

    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    mrid = obj.uri()

    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'

    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <%s>""" % connection_params.namespace

    query_message += """
        SELECT DISTINCT ?identifier ?attribute ?value ?edge
        WHERE {
          """

    query_message += '''
    VALUES ?identifier {"%s"}''' %mrid
    # add all equipment mRID

    query_message += f'''
        bind(iri(concat("{split}", ?identifier)) as ?eq)'''

    query_message += """

        ?eq cim:%s ?val.

        {bind("%s" as ?attribute)}
        {bind(strafter(str(?val),"%s") as ?uri)}
        {bind(if(?uri = "", ?val, ?uri) as ?value)}

        OPTIONAL {?val a ?classraw.
                  bind(strafter(str(?classraw),"%s") as ?edge_class)
                  {bind(strafter(str(?val),"%s") as ?uri)}

                  bind(concat("{\\"@id\\":\\"", ?uri,"\\",\\"@type\\":\\"", ?edge_class, "\\"}") as ?edge)}
        }

        ORDER by  ?identifier ?attribute
        """ % (attribute, attribute, split, connection_params.namespace, split)
    return query_message



# from cimgraph.utils.write_json import write_json_ld
from cimgraph.utils.get_all_data import (get_all_bus_data, get_all_data,
                                         get_all_inverter_data,
                                         get_all_limit_data, get_all_line_data,
                                         get_all_load_data,
                                         get_all_location_data,
                                         get_all_measurement_data,
                                         get_all_switch_data,
                                         get_all_transformer_data)
from cimgraph.utils.object_utils import create_object
from cimgraph.utils.write_csv import write_csv
from cimgraph.utils.write_xml import write_xml



'''
    Annotated CIMantic Graphs data profile init file for cim17v40
    Generated by CIMTool http://cimtool.org
'''
from cimgraph.data_profile.cim17v40.cim17v40 import (
    PU, Accumulator, AccumulatorLimit, AccumulatorLimitSet, AccumulatorReset,
    AccumulatorValue, ACDCConverter, ACDCConverterDCTerminal, ACDCTerminal,
    ACLineSegment, ACLineSegmentPhase, ActivePower, ActivePowerChangeRate,
    ActivePowerLimit, ActivePowerPerCurrentFlow, ActivePowerPerFrequency,
    Analog, AnalogControl, AnalogLimit, AnalogLimitSet, AnalogValue,
    AngleDegrees, AngleRadians, ApparentPower, ApparentPowerLimit, Asset,
    AssetContainer, AssetInfo, AssetKind, AsynchronousMachine,
    AsynchronousMachineDynamics, AsynchronousMachineKind, BaseFrequency,
    BasePower, BaseVoltage, BasicIntervalSchedule, BatteryStateKind,
    BatteryUnit, Bay, BranchGroup, BranchGroupTerminal, Breaker,
    BreakerConfiguration, BusbarConfiguration, BusbarSection,
    BusbarSectionInfo, Bushing, BushingInfo, BushingInsulationKind,
    BusNameMarker, CableConstructionKind, CableInfo, CableOuterJacketKind,
    CableShieldMaterialKind, CalculationMethodHierarchy, Capacitance,
    CapacitancePerLength, CatalogAssetType, ChangeSetMember, Circuit, Clamp,
    ClampAction, Classification, Command, CompositeSwitch,
    ConcentricNeutralCableInfo, Conductance, ConductancePerLength,
    ConductingEquipment, Conductor, ConformLoad, ConformLoadGroup,
    ConformLoadSchedule, ConnectivityNode, ConnectivityNodeContainer,
    Connector, Control, ControlAction, ControlArea, ControlAreaGeneratingUnit,
    CoolantType, CoordinateSystem, CSCDynamics, CsConverter,
    CsOperatingModeKind, CsPpccControlKind, Currency, CurrentFlow,
    CurrentLimit, Curve, CurveData, CurveStyle, Customer, CustomerAgreement,
    CustomerKind, Cut, CutAction, DateInterval, DateTimeInterval, DayType,
    DCBaseTerminal, DCBreaker, DCBusbar, DCChopper, DCConductingEquipment,
    DCConverterOperatingModeKind, DCConverterUnit, DCDisconnector,
    DCEquipmentContainer, DCGround, DCLine, DCLineSegment, DCNode,
    DCPolarityKind, DCSeriesDevice, DCShunt, DCSwitch, DCTerminal,
    DCTopologicalIsland, DCTopologicalNode, DecimalQuantity,
    DemandResponseProgram, DERCurveData, DERFunction, DERGroupDispatch,
    DERGroupForecast, DERMonitorableParameter, DERParameterKind, DERUnitSymbol,
    DisconnectingCircuitBreaker, Disconnector, Discrete, DiscreteValue,
    DispatchSchedule, Displacement, DuctBank, EarthFaultCompensator, EndDevice,
    EndDeviceControl, EndDeviceGroup, EnergyArea, EnergyConnection,
    EnergyConsumer, EnergyConsumerAction, EnergyConsumerPhase,
    EnergySchedulingType, EnergySource, EnergySourceAction, EnergySourcePhase,
    Equipment, EquipmentContainer, ExternalNetworkInjection, Feeder,
    FloatQuantity, FlowDirectionKind, Frequency, FrequencyConverter, Fuse,
    GeneratingUnit, GeneratorControlMode, GeneratorControlSource,
    GenUnitOpSchedule, GeographicalRegion, Ground, GroundAction,
    GroundDisconnector, GroundingImpedance, Hours, HydroPump, IdentifiedObject,
    Identity, IEC61970CIMVersion, Impedance, Inductance, InductancePerLength,
    InstanceSet, IntegerQuantity, InterrupterUnitInfo, InterruptingMediumKind,
    InUseStateKind, IOPoint, IOPointSource, IrregularIntervalSchedule,
    IrregularTimePoint, Jumper, JumperAction, Junction, KiloActivePower,
    Length, Limit, LimitSet, Line, LinearShuntCompensator,
    LinearShuntCompensatorPhase, LoadArea, LoadBreakSwitch, LoadDynamics,
    LoadGroup, LoadResponseCharacteristic, Location, Mass, Measurement,
    MeasurementAction, MeasurementValue, MeasurementValueQuality,
    MeasurementValueSource, MeterReadSchedule, Minutes, Money, MonthDay,
    MonthDayInterval, MutualCoupling, Name, NameType, NameTypeAuthority,
    NoLoadTest, NonConformLoad, NonConformLoadGroup, NonConformLoadSchedule,
    NonlinearShuntCompensator, NonlinearShuntCompensatorPhase,
    NonlinearShuntCompensatorPhasePoint, NonlinearShuntCompensatorPoint,
    OldPerson, OpenCircuitTest, OperatingMechanismInfo, OperatingMechanismKind,
    OperatingParticipant, OperatingShare, OperationalLimit,
    OperationalLimitDirectionKind, OperationalLimitSet, OperationalLimitType,
    OperationalRestriction, OperatonalLimitTypeScaling, Outage,
    OverheadWireInfo, PerCent, PerLengthDCLineParameter, PerLengthImpedance,
    PerLengthLineParameter, PerLengthPhaseImpedance,
    PerLengthSequenceImpedance, PetersenCoil, PetersenCoilModeKind, PhaseCode,
    PhaseImpedanceData, PhaseShuntConnectionKind, PhaseTapChanger,
    PhaseTapChangerAsymmetrical, PhaseTapChangerLinear,
    PhaseTapChangerNonLinear, PhaseTapChangerSymmetrical, PhaseTapChangerTable,
    PhaseTapChangerTablePoint, PhaseTapChangerTabular, PhotoVoltaicUnit, Plant,
    PositionPoint, PowerCutZone, PowerElectronicsConnection,
    PowerElectronicsConnectionPhase, PowerElectronicsUnit,
    PowerElectronicsWindUnit, PowerSystemResource, PowerTransformer,
    PowerTransformerEnd, PowerTransformerInfo, Pressure, ProductAssetModel,
    ProtectedSwitch, ProtectiveActionAdjustment, PSRType, Quality61850,
    RaiseLowerCommand, RatioTapChanger, RatioTapChangerTable,
    RatioTapChangerTablePoint, Reactance, ReactancePerLength,
    ReactiveCapabilityCurve, ReactivePower, RealEnergy, Recloser,
    RegularIntervalSchedule, RegularTimePoint, RegulatingCondEq,
    RegulatingControl, RegulatingControlModeKind, RegulationBranchKind,
    RegulationSchedule, RemoteControl, RemoteSource, ReportingGroup,
    ReportingSuperGroup, Resistance, ResistancePerLength, RotatingMachine,
    RotationSpeed, Season, SeasonDayTypeSchedule, Seconds, Sectionaliser,
    SeriesCompensator, SetPoint, ShortCircuitRotorKind, ShortCircuitTest,
    ShuntCompensator, ShuntCompensatorAction, ShuntCompensatorControl,
    ShuntCompensatorInfo, ShuntCompensatorPhase, ShuntImpedanceControlKind,
    ShuntImpedanceLocalControlKind, SinglePhaseKind, Source,
    StaticVarCompensator, StaticVarCompensatorDynamics, StationSupply, Status,
    StringMeasurement, StringMeasurementValue, StringQuantity,
    SubGeographicalRegion, SubLoadArea, Substation, Susceptance,
    SusceptancePerLength, SVCControlMode, SvShuntCompensatorSections,
    SvTapStep, Switch, SwitchAction, SwitchActionKind, SwitchInfo, SwitchPhase,
    SwitchSchedule, SynchronousMachine, SynchronousMachineDynamics,
    SynchronousMachineKind, SynchronousMachineOperatingMode, TapChanger,
    TapChangerControl, TapChangerInfo, TapChangerTablePoint,
    TapeShieldCableInfo, TapSchedule, TempEquipActionKind, Temperature,
    Terminal, TimeInterval, TimeIntervalKind, TopologicalIsland,
    TopologicalNode, TransformerControlMode, TransformerCoreAdmittance,
    TransformerEnd, TransformerEndInfo, TransformerMeshImpedance,
    TransformerObservation, TransformerStarImpedance, TransformerTank,
    TransformerTankEnd, TransformerTankInfo, TransformerTest, UnitMultiplier,
    UnitSymbol, UsagePointGroup, Validity, ValueAliasSet, ValueToAlias,
    Version, Voltage, VoltageControlZone, VoltageLevel, VoltageLimit,
    VoltagePerReactivePower, Volume, VsCapabilityCurve, VSCDynamics,
    VsConverter, VsPpccControlKind, VsQpccControlKind, WindGeneratingUnit,
    WindGenUnitKind, WindingConnection, WindingInsulation,
    WindTurbineType3or4Dynamics, WireAssemblyInfo, WireInfo,
    WireInsulationKind, WireMaterialKind, WirePhaseInfo, WirePosition,
    WireSegment, WireSegmentPhase, WireSpacingInfo, WireUsageKind)

__all__ = [
    'Identity',
    'ACDCConverter' ,
    'ACDCConverterDCTerminal' ,
    'ACDCTerminal' ,
    'ACLineSegment' ,
    'ACLineSegmentPhase' ,
    'Accumulator' ,
    'AccumulatorLimit' ,
    'AccumulatorLimitSet' ,
    'AccumulatorReset' ,
    'AccumulatorValue' ,
    'ActivePowerLimit' ,
    'Analog' ,
    'AnalogControl' ,
    'AnalogLimit' ,
    'AnalogLimitSet' ,
    'AnalogValue' ,
    'ApparentPowerLimit' ,
    'Asset' ,
    'AssetContainer' ,
    'AssetInfo' ,
    'AssetKind' ,
    'AsynchronousMachine' ,
    'AsynchronousMachineDynamics' ,
    'AsynchronousMachineKind' ,
    'BaseFrequency' ,
    'BasePower' ,
    'BaseVoltage' ,
    'BasicIntervalSchedule' ,
    'BatteryStateKind' ,
    'BatteryUnit' ,
    'Bay' ,
    'BranchGroup' ,
    'BranchGroupTerminal' ,
    'Breaker' ,
    'BreakerConfiguration' ,
    'BusNameMarker' ,
    'BusbarConfiguration' ,
    'BusbarSection' ,
    'BusbarSectionInfo' ,
    'Bushing' ,
    'BushingInfo' ,
    'BushingInsulationKind' ,
    'CSCDynamics' ,
    'CableConstructionKind' ,
    'CableInfo' ,
    'CableOuterJacketKind' ,
    'CableShieldMaterialKind' ,
    'CalculationMethodHierarchy' ,
    'CatalogAssetType' ,
    'ChangeSetMember' ,
    'Circuit' ,
    'Clamp' ,
    'ClampAction' ,
    'Command' ,
    'CompositeSwitch' ,
    'ConcentricNeutralCableInfo' ,
    'ConductingEquipment' ,
    'Conductor' ,
    'ConformLoad' ,
    'ConformLoadGroup' ,
    'ConformLoadSchedule' ,
    'ConnectivityNode' ,
    'ConnectivityNodeContainer' ,
    'Connector' ,
    'Control' ,
    'ControlAction' ,
    'ControlArea' ,
    'ControlAreaGeneratingUnit' ,
    'CoolantType' ,
    'CoordinateSystem' ,
    'CsConverter' ,
    'CsOperatingModeKind' ,
    'CsPpccControlKind' ,
    'Currency' ,
    'CurrentLimit' ,
    'Curve' ,
    'CurveData' ,
    'CurveStyle' ,
    'Customer' ,
    'CustomerAgreement' ,
    'CustomerKind' ,
    'Cut' ,
    'CutAction' ,
    'DCBaseTerminal' ,
    'DCBreaker' ,
    'DCBusbar' ,
    'DCChopper' ,
    'DCConductingEquipment' ,
    'DCConverterOperatingModeKind' ,
    'DCConverterUnit' ,
    'DCDisconnector' ,
    'DCEquipmentContainer' ,
    'DCGround' ,
    'DCLine' ,
    'DCLineSegment' ,
    'DCNode' ,
    'DCPolarityKind' ,
    'DCSeriesDevice' ,
    'DCShunt' ,
    'DCSwitch' ,
    'DCTerminal' ,
    'DCTopologicalIsland' ,
    'DCTopologicalNode' ,
    'DERCurveData' ,
    'DERFunction' ,
    'DERGroupDispatch' ,
    'DERGroupForecast' ,
    'DERMonitorableParameter' ,
    'DERParameterKind' ,
    'DERUnitSymbol' ,
    'DateInterval' ,
    'DateTimeInterval' ,
    'DayType' ,
    'DecimalQuantity' ,
    'DemandResponseProgram' ,
    'DisconnectingCircuitBreaker' ,
    'Disconnector' ,
    'Discrete' ,
    'DiscreteValue' ,
    'DispatchSchedule' ,
    'DuctBank' ,
    'EarthFaultCompensator' ,
    'EndDevice' ,
    'EndDeviceControl' ,
    'EndDeviceGroup' ,
    'EnergyArea' ,
    'EnergyConnection' ,
    'EnergyConsumer' ,
    'EnergyConsumerAction' ,
    'EnergyConsumerPhase' ,
    'EnergySchedulingType' ,
    'EnergySource' ,
    'EnergySourceAction' ,
    'EnergySourcePhase' ,
    'Equipment' ,
    'EquipmentContainer' ,
    'ExternalNetworkInjection' ,
    'Feeder' ,
    'FloatQuantity' ,
    'FlowDirectionKind' ,
    'FrequencyConverter' ,
    'Fuse' ,
    'GenUnitOpSchedule' ,
    'GeneratingUnit' ,
    'GeneratorControlMode' ,
    'GeneratorControlSource' ,
    'GeographicalRegion' ,
    'Ground' ,
    'GroundAction' ,
    'GroundDisconnector' ,
    'GroundingImpedance' ,
    'HydroPump' ,
    'IEC61970CIMVersion' ,
    'IOPoint' ,
    'IOPointSource' ,
    'IdentifiedObject' ,
    'InUseStateKind' ,
    'InstanceSet' ,
    'IntegerQuantity' ,
    'InterrupterUnitInfo' ,
    'InterruptingMediumKind' ,
    'IrregularIntervalSchedule' ,
    'IrregularTimePoint' ,
    'Jumper' ,
    'JumperAction' ,
    'Junction' ,
    'Limit' ,
    'LimitSet' ,
    'Line' ,
    'LinearShuntCompensator' ,
    'LinearShuntCompensatorPhase' ,
    'LoadArea' ,
    'LoadBreakSwitch' ,
    'LoadDynamics' ,
    'LoadGroup' ,
    'LoadResponseCharacteristic' ,
    'Location' ,
    'Measurement' ,
    'MeasurementAction' ,
    'MeasurementValue' ,
    'MeasurementValueQuality' ,
    'MeasurementValueSource' ,
    'MeterReadSchedule' ,
    'MonthDayInterval' ,
    'MutualCoupling' ,
    'Name' ,
    'NameType' ,
    'NameTypeAuthority' ,
    'NoLoadTest' ,
    'NonConformLoad' ,
    'NonConformLoadGroup' ,
    'NonConformLoadSchedule' ,
    'NonlinearShuntCompensator' ,
    'NonlinearShuntCompensatorPhase' ,
    'NonlinearShuntCompensatorPhasePoint' ,
    'NonlinearShuntCompensatorPoint' ,
    'OldPerson' ,
    'OpenCircuitTest' ,
    'OperatingMechanismInfo' ,
    'OperatingMechanismKind' ,
    'OperatingParticipant' ,
    'OperatingShare' ,
    'OperationalLimit' ,
    'OperationalLimitDirectionKind' ,
    'OperationalLimitSet' ,
    'OperationalLimitType' ,
    'OperationalRestriction' ,
    'OperatonalLimitTypeScaling' ,
    'Outage' ,
    'OverheadWireInfo' ,
    'PSRType' ,
    'PerLengthDCLineParameter' ,
    'PerLengthImpedance' ,
    'PerLengthLineParameter' ,
    'PerLengthPhaseImpedance' ,
    'PerLengthSequenceImpedance' ,
    'PetersenCoil' ,
    'PetersenCoilModeKind' ,
    'PhaseCode' ,
    'PhaseImpedanceData' ,
    'PhaseShuntConnectionKind' ,
    'PhaseTapChanger' ,
    'PhaseTapChangerAsymmetrical' ,
    'PhaseTapChangerLinear' ,
    'PhaseTapChangerNonLinear' ,
    'PhaseTapChangerSymmetrical' ,
    'PhaseTapChangerTable' ,
    'PhaseTapChangerTablePoint' ,
    'PhaseTapChangerTabular' ,
    'PhotoVoltaicUnit' ,
    'Plant' ,
    'PositionPoint' ,
    'PowerCutZone' ,
    'PowerElectronicsConnection' ,
    'PowerElectronicsConnectionPhase' ,
    'PowerElectronicsUnit' ,
    'PowerElectronicsWindUnit' ,
    'PowerSystemResource' ,
    'PowerTransformer' ,
    'PowerTransformerEnd' ,
    'PowerTransformerInfo' ,
    'ProductAssetModel' ,
    'ProtectedSwitch' ,
    'ProtectiveActionAdjustment' ,
    'Quality61850' ,
    'RaiseLowerCommand' ,
    'RatioTapChanger' ,
    'RatioTapChangerTable' ,
    'RatioTapChangerTablePoint' ,
    'ReactiveCapabilityCurve' ,
    'Recloser' ,
    'RegularIntervalSchedule' ,
    'RegularTimePoint' ,
    'RegulatingCondEq' ,
    'RegulatingControl' ,
    'RegulatingControlModeKind' ,
    'RegulationBranchKind' ,
    'RegulationSchedule' ,
    'RemoteControl' ,
    'RemoteSource' ,
    'ReportingGroup' ,
    'ReportingSuperGroup' ,
    'RotatingMachine' ,
    'SVCControlMode' ,
    'Season' ,
    'SeasonDayTypeSchedule' ,
    'Sectionaliser' ,
    'SeriesCompensator' ,
    'SetPoint' ,
    'ShortCircuitRotorKind' ,
    'ShortCircuitTest' ,
    'ShuntCompensator' ,
    'ShuntCompensatorAction' ,
    'ShuntCompensatorControl' ,
    'ShuntCompensatorInfo' ,
    'ShuntCompensatorPhase' ,
    'ShuntImpedanceControlKind' ,
    'ShuntImpedanceLocalControlKind' ,
    'SinglePhaseKind' ,
    'Source' ,
    'StaticVarCompensator' ,
    'StaticVarCompensatorDynamics' ,
    'StationSupply' ,
    'Status' ,
    'StringMeasurement' ,
    'StringMeasurementValue' ,
    'StringQuantity' ,
    'SubGeographicalRegion' ,
    'SubLoadArea' ,
    'Substation' ,
    'SvShuntCompensatorSections' ,
    'SvTapStep' ,
    'Switch' ,
    'SwitchAction' ,
    'SwitchActionKind' ,
    'SwitchInfo' ,
    'SwitchPhase' ,
    'SwitchSchedule' ,
    'SynchronousMachine' ,
    'SynchronousMachineDynamics' ,
    'SynchronousMachineKind' ,
    'SynchronousMachineOperatingMode' ,
    'TapChanger' ,
    'TapChangerControl' ,
    'TapChangerInfo' ,
    'TapChangerTablePoint' ,
    'TapSchedule' ,
    'TapeShieldCableInfo' ,
    'TempEquipActionKind' ,
    'Terminal' ,
    'TimeInterval' ,
    'TimeIntervalKind' ,
    'TopologicalIsland' ,
    'TopologicalNode' ,
    'TransformerControlMode' ,
    'TransformerCoreAdmittance' ,
    'TransformerEnd' ,
    'TransformerEndInfo' ,
    'TransformerMeshImpedance' ,
    'TransformerObservation' ,
    'TransformerStarImpedance' ,
    'TransformerTank' ,
    'TransformerTankEnd' ,
    'TransformerTankInfo' ,
    'TransformerTest' ,
    'UnitMultiplier' ,
    'UnitSymbol' ,
    'UsagePointGroup' ,
    'VSCDynamics' ,
    'Validity' ,
    'ValueAliasSet' ,
    'ValueToAlias' ,
    'Version' ,
    'VoltageControlZone' ,
    'VoltageLevel' ,
    'VoltageLimit' ,
    'VsCapabilityCurve' ,
    'VsConverter' ,
    'VsPpccControlKind' ,
    'VsQpccControlKind' ,
    'WindGenUnitKind' ,
    'WindGeneratingUnit' ,
    'WindTurbineType3or4Dynamics' ,
    'WindingConnection' ,
    'WindingInsulation' ,
    'WireAssemblyInfo' ,
    'WireInfo' ,
    'WireInsulationKind' ,
    'WireMaterialKind' ,
    'WirePhaseInfo' ,
    'WirePosition' ,
    'WireSegment' ,
    'WireSegmentPhase' ,
    'WireSpacingInfo' ,
    'WireUsageKind' ,
    'MonthDay' ,
    'ReactivePower' ,
    'Susceptance' ,
    'Pressure' ,
    'Hours' ,
    'Capacitance' ,
    'Temperature' ,
    'Seconds' ,
    'Voltage' ,
    'ActivePowerPerFrequency' ,
    'Inductance' ,
    'CapacitancePerLength' ,
    'SusceptancePerLength' ,
    'Frequency' ,
    'PU' ,
    'VoltagePerReactivePower' ,
    'ActivePower' ,
    'Mass' ,
    'ConductancePerLength' ,
    'Reactance' ,
    'Minutes' ,
    'AngleDegrees' ,
    'ActivePowerChangeRate' ,
    'RotationSpeed' ,
    'RealEnergy' ,
    'Conductance' ,
    'ReactancePerLength' ,
    'Resistance' ,
    'Money' ,
    'Classification' ,
    'ActivePowerPerCurrentFlow' ,
    'Volume' ,
    'Length' ,
    'KiloActivePower' ,
    'ResistancePerLength' ,
    'InductancePerLength' ,
    'Displacement' ,
    'ApparentPower' ,
    'PerCent' ,
    'AngleRadians' ,
    'CurrentFlow' ,
    'Impedance'
]



from __future__ import annotations

from cimgraph.databases import ConnectionParameters


def get_all_nodes_from_container(container: object, namespace: str):
    """
    Generates SPARQL query string for all nodes, terminals, and conducting equipment
    Args:

    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    container_class = container.__class__.__name__
    uri = container.uri()

    query_message = f"""MATCH (container:{container_class})
WHERE container.uri = "{uri}"
MATCH (eq) - [:`Equipment.EquipmentContainer`] - (container)
OPTIONAL MATCH (cnode) - [:`Terminal.ConnectivityNode`] - (term:Terminal) - [:`Terminal.ConductingEquipment`] -> (eq)
RETURN DISTINCT
cnode.`IdentifiedObject.mRID` as ConnectivityNode,
term.`IdentifiedObject.mRID` as Terminal,
eq.`IdentifiedObject.mRID` as eq_id,
LABELS(eq)[1] as eq_class"""

    return query_message


def get_all_nodes_from_area(container: object, namespace: str):
    """
    Generates SPARQL query string for all nodes, terminals, and conducting equipment
    Args:

    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    container_class = container.__class__.__name__
    uri = container.uri()

    query_message = f"""MATCH (container:{container_class})
WHERE container.uri = "{uri}"
MATCH (eq) - [:`Equipment.SubSchedulingArea`] - (container)
OPTIONAL MATCH (cnode) - [:`Terminal.ConnectivityNode`] - (term:Terminal) - [:`Terminal.ConductingEquipment`] -> (eq)
RETURN DISTINCT
cnode.`IdentifiedObject.mRID` as ConnectivityNode,
term.`IdentifiedObject.mRID` as Terminal,
eq.`IdentifiedObject.mRID` as eq_id,
LABELS(eq)[1] as eq_class"""

    return query_message



'''
    Annotated CIMantic Graphs data profile init file for Profile
    Generated by CIMTool http://cimtool.org
'''
from cimgraph.data_profile.rc4_2021.rc4_2021 import (
    PU, AccountNotification, AccumulationKind, Accumulator, AccumulatorLimit,
    AccumulatorLimitSet, AccumulatorReset, AccumulatorValue, ACDCTerminal,
    ACLineSegment, ACLineSegmentPhase, ActivePower, ActivePowerChangeRate,
    ActivePowerLimit, ActivePowerPerFrequency, ActivityRecord, AggregateKind,
    AggregateScore, AirCompressor, AlertTypeList, AmiBillingReadyKind, Analog,
    AnalogControl, AnalogLimit, AnalogLimitSet, AnalogValue, Analytic,
    AnalyticKind, AnalyticScore, AnchorKind, AngleDegrees, AngleRadians,
    ApparentPower, ApparentPowerLimit, Appointment, Approver, Area, Asset,
    AssetContainer, AssetDeployment, AssetFunction, AssetGroup, AssetGroupKind,
    AssetHazardKind, AssetHealthEvent, AssetInfo, AssetKind,
    AssetLocationHazard, AssetModelUsageKind, AssetOrganisationRole,
    AssetOwner, AsynchronousMachine, AsynchronousMachineKind, Author,
    BaseFrequency, BasePower, BaseVoltage, BasicIntervalSchedule, BatteryState,
    BatteryUnit, Bay, BranchGroup, BranchGroupTerminal, Breaker,
    BreakerApplicationKind, BreakerConfiguration, BusbarConfiguration,
    BusbarSection, BusbarSectionInfo, Bushing, BushingInfo,
    BushingInsulationKind, BusNameMarker, CableConstructionKind, CableInfo,
    CableOuterJacketKind, CableShieldMaterialKind, CAESPlant,
    CalculationIntervalUnitKind, Capacitance, CatalogAssetType, Channel, Clamp,
    Class1, ClassificationCondition, CogenerationPlant, CombinedCyclePlant,
    Command, CommodityKind, CompositeSwitch, ConcentricNeutralCableInfo,
    Conductance, ConductancePerLength, ConductingEquipment, Conductor,
    ConfigurationEvent, ConformLoad, ConformLoadGroup, ConformLoadSchedule,
    ConnectivityNode, ConnectivityNodeContainer, Connector, Control,
    CoolantType, CoordinateSystem, CorporateStandardKind, CostPerEnergyUnit,
    CostPerHeatUnit, CostPerVolume, CostRate, Crew, CrewMember, CrewStatusKind,
    CrewType, Currency, CurrentFlow, CurrentLimit, Curve, CurveData,
    CurveStyle, Customer, CustomerAccount, CustomerAgreement, CustomerKind,
    CustomerNotification, Cut, DateInterval, DateTimeInterval, DayType,
    DecimalQuantity, DemandResponseProgram, DeploymentDate,
    DeploymentStateKind, DERCurveData, DERFunction, DERGroupDispatch,
    DERGroupForecast, DERMonitorableParameter, DERParameterKind, DERUnitSymbol,
    Disconnector, Discrete, DiscreteCommand, DiscreteValue,
    DispatchablePowerCapability, DispatchSchedule, Displacement, Document,
    DuctBank, EarthFaultCompensator, Editor, ElectronicAddress, Emission,
    EmissionAccount, EmissionCurve, EmissionType, EmissionValueSource,
    EndDevice, EndDeviceAction, EndDeviceCapability, EndDeviceControl,
    EndDeviceControlType, EndDeviceEvent, EndDeviceEventDetail,
    EndDeviceEventType, EndDeviceFunction, EndDeviceGroup, EndDeviceInfo,
    EndDeviceTiming, EnergyArea, EnergyConnection, EnergyConsumer,
    EnergyConsumerPhase, EnergySource, EnergySourcePhase, EnvironmentalAlert,
    EnvironmentalAnalog, EnvironmentalDataAuthority, EnvironmentalDataProvider,
    EnvironmentalInformation, EnvironmentalLocationType,
    EnvironmentalPhenomenon, EnvironmentalStringMeasurement, Equipment,
    EquipmentContainer, EquipmentFault, Estimate, ExternalNetworkInjection,
    FacilityKind, Fault, FaultCauseType, FaultImpedance, Feeder,
    FieldDispatchHistory, FieldDispatchStep, FinancialInfo, FloatQuantity,
    FlowDirectionKind, FossilFuel, Frequency, FrequencyConverter,
    FuelAllocationSchedule, FuelType, Fuse, GeneratingUnit,
    GeneratorControlMode, GeneratorControlSource, GenUnitOpCostCurve,
    GenUnitOpSchedule, GeographicalRegion, GrossToNetActivePowerCurve, Ground,
    GroundDisconnector, GroundingImpedance, HeatInputCurve, HeatRate,
    HeatRateCurve, Hours, House, HouseCooling, HouseHeating,
    HouseThermalIntegrity, HydroEnergyConversionKind,
    HydroGeneratingEfficiencyCurve, HydroGeneratingUnit, HydroPlantStorageKind,
    HydroPowerPlant, HydroPump, HydroPumpOpSchedule, IdentifiedObject,
    Identity, IEC61970CIMVersion, IEEE1547AbnormalPerfomanceCategory,
    IEEE1547ControlSettings, IEEE1547Info, IEEE1547IslandingCategory,
    IEEE1547NormalPerformanceCategory, IEEE1547Setting, IEEE1547TripSettings,
    Impedance, IncrementalHeatRateCurve, InflowForecast, InspectionDataSet,
    IntegerQuantity, InterrupterUnit, InterrupterUnitInfo,
    InterruptingMediumKind, IntervalBlock, IntervalReading, IOPoint,
    IrregularIntervalSchedule, IrregularTimePoint, Issuer, Jumper, Junction,
    KiloActivePower, Length, LevelVsVolumeCurve, Limit, LimitSet, Line,
    LinearShuntCompensator, LinearShuntCompensatorPhase, LineFault, LoadArea,
    LoadBreakSwitch, LoadGroup, LoadResponseCharacteristic, Location,
    LocationKind, MacroPeriodKind, Manufacturer, Mass, Measurement,
    MeasurementKind, MeasurementValue, MeasurementValueQuality,
    MeasurementValueSource, MeasuringPeriodKind, Medium, MediumKind, Meter,
    MeterMultiplier, MeterMultiplierKind, MeterReading, MeterWorkTask,
    MetrologyRequirement, Minutes, Money, MonthDay, MonthDayInterval,
    MutualCoupling, Name, NamedPhenomenon, NameType, NameTypeAuthority,
    NoLoadTest, NonConformLoad, NonConformLoadGroup, NonConformLoadSchedule,
    NonlinearShuntCompensator, NonlinearShuntCompensatorPhase,
    NonlinearShuntCompensatorPhasePoint, NonlinearShuntCompensatorPoint,
    NotificationTriggerKind, NuclearGeneratingUnit, OpenCircuitTest,
    OperatingMechanism, OperatingMechanismInfo, OperatingMechanismKind,
    OperatingParticipant, OperatingShare, OperationalLimit,
    OperationalLimitDirectionKind, OperationalLimitSet, OperationalLimitType,
    Organisation, OrganisationRole, OverfrequencyTripCurve, OverheadWireInfo,
    OvervoltageTripCurve, Ownership, ParallelLineSegment, ParentOrganization,
    PendingCalculation, PenstockLossCurve, PerCent, PerLengthImpedance,
    PerLengthLineParameter, PerLengthPhaseImpedance,
    PerLengthSequenceImpedance, Person, PersonRole, PetersenCoil,
    PetersenCoilModeKind, PhaseCode, PhaseConnectedFaultKind,
    PhaseImpedanceData, PhaseShuntConnectionKind, PhaseTapChanger,
    PhaseTapChangerAsymmetrical, PhaseTapChangerLinear,
    PhaseTapChangerNonLinear, PhaseTapChangerSymmetrical, PhaseTapChangerTable,
    PhaseTapChangerTablePoint, PhaseTapChangerTabular,
    PhenomenonClassification, PhotovoltaicUnit, Plant, PositionPoint,
    PowerCutZone, PowerElectronicsConnection, PowerElectronicsConnectionPhase,
    PowerElectronicsUnit, PowerElectronicsWindUnit, PowerSystemResource,
    PowerTransformer, PowerTransformerEnd, PowerTransformerInfo, Pressure,
    PricingStructure, Priority, Procedure, ProcedureDataSet, ProcedureKind,
    ProductAssetModel, ProtectedSwitch, PSRType, Quality61850,
    RaiseLowerCommand, RandomisationKind, RationalNumber, RatioTapChanger,
    RatioTapChangerTable, RatioTapChangerTablePoint, Reactance,
    ReactancePerLength, ReactiveCapabilityCurve, ReactivePower, Reading,
    ReadingInterharmonic, ReadingReasonKind, ReadingType, RealEnergy, Recloser,
    Register, RegularIntervalSchedule, RegularTimePoint, RegulatingCondEq,
    RegulatingControl, RegulatingControlModeKind, RegulationSchedule,
    ReportingCapability, ReportingGroup, ReportingMethodKind,
    ReportingSuperGroup, Reservoir, Resistance, ResistancePerLength,
    RightOfWay, RotatingMachine, RotationSpeed, ScaleKind, ScheduledEvent,
    ScheduledEventData, Seal, SealConditionKind, SealKind, Season,
    SeasonDayTypeSchedule, Seconds, Sectionaliser, SeriesCompensator,
    ServiceCategory, ServiceKind, ServiceLocation, ServiceMultiplier,
    ServiceMultiplierKind, SetPoint, ShortCircuitRotorKind, ShortCircuitTest,
    ShuntCompensator, ShuntCompensatorInfo, ShuntCompensatorPhase,
    ShutdownCurve, SinglePhaseKind, SmartInverterMode, SolarGeneratingUnit,
    Source, StartIgnFuelCurve, StartMainFuelCurve, StartRampCurve,
    StartupModel, StateVariable, StaticVarCompensator, StationSupply, Status,
    SteamSendoutSchedule, StreetAddress, StreetDetail, StringMeasurement,
    StringMeasurementValue, StringQuantity, Structure, StructureMaterialKind,
    StructureSupport, StructureSupportKind, SubGeographicalRegion, SubLoadArea,
    Substation, Susceptance, SusceptancePerLength, SVCControlMode,
    SvEstVoltage, SvInjection, SvPowerFlow, SvShuntCompensatorSections,
    SvStatus, SvSwitch, SvTapStep, SvVoltage, Switch, SwitchInfo,
    SwitchOperationSummary, SwitchPhase, SwitchSchedule, SynchronousMachine,
    SynchronousMachineKind, SynchronousMachineOperatingMode, TailbayLossCurve,
    TapChanger, TapChangerControl, TapChangerInfo, TapChangerTablePoint,
    TapeShieldCableInfo, TapSchedule, TargetLevelSchedule, TelephoneNumber,
    Temperature, Terminal, TestKind, ThermalGeneratingUnit,
    ThermostatController, ThermostatControlMode, TimeInterval,
    TimeIntervalKind, TopologicalIsland, TopologicalNode, TownDetail,
    TransformerApplicationKind, TransformerControlMode,
    TransformerCoreAdmittance, TransformerEnd, TransformerEndInfo,
    TransformerMeshImpedance, TransformerStarImpedance, TransformerTank,
    TransformerTankEnd, TransformerTankInfo, TransformerTest, TroubleTicket,
    UnderfrequencyTripCurve, UndervoltageTripCurve, UnitMultiplier, UnitSymbol,
    UsagePoint, UsagePointConnectedKind, UsagePointGroup, UsagePointLocation,
    UserAttribute, Validity, ValueAliasSet, ValueToAlias, Version, Voltage,
    VoltageControlZone, VoltageLevel, VoltageLimit, VoltagePerReactivePower,
    VoltVarCurve, VoltWattCurve, Volume, VolumeFlowRate, WaterLevel,
    WattVarCurve, WindGeneratingUnit, WindGenUnitKind, WindingConnection,
    WireAssemblyInfo, WireInfo, WireInsulationKind, WireMaterialKind,
    WirePhaseInfo, WirePosition, WireSpacingInfo, WireUsageKind)

__all__ = [
    'Identity',
    'ACDCTerminal' ,
    'ACLineSegment' ,
    'ACLineSegmentPhase' ,
    'AccountNotification' ,
    'AccumulationKind' ,
    'Accumulator' ,
    'AccumulatorLimit' ,
    'AccumulatorLimitSet' ,
    'AccumulatorReset' ,
    'AccumulatorValue' ,
    'ActivePowerLimit' ,
    'ActivityRecord' ,
    'AggregateKind' ,
    'AggregateScore' ,
    'AirCompressor' ,
    'AlertTypeList' ,
    'AmiBillingReadyKind' ,
    'Analog' ,
    'AnalogControl' ,
    'AnalogLimit' ,
    'AnalogLimitSet' ,
    'AnalogValue' ,
    'Analytic' ,
    'AnalyticKind' ,
    'AnalyticScore' ,
    'ApparentPowerLimit' ,
    'Appointment' ,
    'Approver' ,
    'Asset' ,
    'AssetContainer' ,
    'AssetDeployment' ,
    'AssetFunction' ,
    'AssetGroup' ,
    'AssetGroupKind' ,
    'AssetHazardKind' ,
    'AssetHealthEvent' ,
    'AssetInfo' ,
    'AssetKind' ,
    'AssetLocationHazard' ,
    'AssetModelUsageKind' ,
    'AssetOrganisationRole' ,
    'AssetOwner' ,
    'AsynchronousMachine' ,
    'AsynchronousMachineKind' ,
    'Author' ,
    'BaseFrequency' ,
    'BasePower' ,
    'BaseVoltage' ,
    'BasicIntervalSchedule' ,
    'BatteryState' ,
    'BatteryUnit' ,
    'Bay' ,
    'BranchGroup' ,
    'BranchGroupTerminal' ,
    'Breaker' ,
    'BreakerApplicationKind' ,
    'BreakerConfiguration' ,
    'BusNameMarker' ,
    'BusbarConfiguration' ,
    'BusbarSection' ,
    'BusbarSectionInfo' ,
    'Bushing' ,
    'BushingInfo' ,
    'BushingInsulationKind' ,
    'CAESPlant' ,
    'CableConstructionKind' ,
    'CableInfo' ,
    'CableOuterJacketKind' ,
    'CableShieldMaterialKind' ,
    'CatalogAssetType' ,
    'Channel' ,
    'Clamp' ,
    'Class1' ,
    'ClassificationCondition' ,
    'CogenerationPlant' ,
    'CombinedCyclePlant' ,
    'Command' ,
    'CommodityKind' ,
    'CompositeSwitch' ,
    'ConcentricNeutralCableInfo' ,
    'ConductingEquipment' ,
    'Conductor' ,
    'ConfigurationEvent' ,
    'ConformLoad' ,
    'ConformLoadGroup' ,
    'ConformLoadSchedule' ,
    'ConnectivityNode' ,
    'ConnectivityNodeContainer' ,
    'Connector' ,
    'Control' ,
    'CoolantType' ,
    'CoordinateSystem' ,
    'CorporateStandardKind' ,
    'Crew' ,
    'CrewMember' ,
    'CrewStatusKind' ,
    'CrewType' ,
    'Currency' ,
    'CurrentLimit' ,
    'Curve' ,
    'CurveData' ,
    'CurveStyle' ,
    'Customer' ,
    'CustomerAccount' ,
    'CustomerAgreement' ,
    'CustomerKind' ,
    'CustomerNotification' ,
    'Cut' ,
    'DERCurveData' ,
    'DERFunction' ,
    'DERGroupDispatch' ,
    'DERGroupForecast' ,
    'DERMonitorableParameter' ,
    'DERParameterKind' ,
    'DERUnitSymbol' ,
    'DateInterval' ,
    'DateTimeInterval' ,
    'DayType' ,
    'DecimalQuantity' ,
    'DemandResponseProgram' ,
    'DeploymentDate' ,
    'DeploymentStateKind' ,
    'Disconnector' ,
    'Discrete' ,
    'DiscreteCommand' ,
    'DiscreteValue' ,
    'DispatchSchedule' ,
    'DispatchablePowerCapability' ,
    'Document' ,
    'DuctBank' ,
    'EarthFaultCompensator' ,
    'Editor' ,
    'ElectronicAddress' ,
    'EmissionAccount' ,
    'EmissionCurve' ,
    'EmissionType' ,
    'EmissionValueSource' ,
    'EndDevice' ,
    'EndDeviceAction' ,
    'EndDeviceCapability' ,
    'EndDeviceControl' ,
    'EndDeviceControlType' ,
    'EndDeviceEvent' ,
    'EndDeviceEventDetail' ,
    'EndDeviceEventType' ,
    'EndDeviceFunction' ,
    'EndDeviceGroup' ,
    'EndDeviceInfo' ,
    'EndDeviceTiming' ,
    'EnergyArea' ,
    'EnergyConnection' ,
    'EnergyConsumer' ,
    'EnergyConsumerPhase' ,
    'EnergySource' ,
    'EnergySourcePhase' ,
    'EnvironmentalAlert' ,
    'EnvironmentalAnalog' ,
    'EnvironmentalDataAuthority' ,
    'EnvironmentalDataProvider' ,
    'EnvironmentalInformation' ,
    'EnvironmentalLocationType' ,
    'EnvironmentalPhenomenon' ,
    'EnvironmentalStringMeasurement' ,
    'Equipment' ,
    'EquipmentContainer' ,
    'EquipmentFault' ,
    'Estimate' ,
    'ExternalNetworkInjection' ,
    'FacilityKind' ,
    'Fault' ,
    'FaultCauseType' ,
    'FaultImpedance' ,
    'Feeder' ,
    'FieldDispatchHistory' ,
    'FieldDispatchStep' ,
    'FinancialInfo' ,
    'FloatQuantity' ,
    'FlowDirectionKind' ,
    'FossilFuel' ,
    'FrequencyConverter' ,
    'FuelAllocationSchedule' ,
    'FuelType' ,
    'Fuse' ,
    'GenUnitOpCostCurve' ,
    'GenUnitOpSchedule' ,
    'GeneratingUnit' ,
    'GeneratorControlMode' ,
    'GeneratorControlSource' ,
    'GeographicalRegion' ,
    'GrossToNetActivePowerCurve' ,
    'Ground' ,
    'GroundDisconnector' ,
    'GroundingImpedance' ,
    'HeatInputCurve' ,
    'HeatRateCurve' ,
    'House' ,
    'HouseCooling' ,
    'HouseHeating' ,
    'HouseThermalIntegrity' ,
    'HydroEnergyConversionKind' ,
    'HydroGeneratingEfficiencyCurve' ,
    'HydroGeneratingUnit' ,
    'HydroPlantStorageKind' ,
    'HydroPowerPlant' ,
    'HydroPump' ,
    'HydroPumpOpSchedule' ,
    'IEC61970CIMVersion' ,
    'IEEE1547AbnormalPerfomanceCategory' ,
    'IEEE1547ControlSettings' ,
    'IEEE1547Info' ,
    'IEEE1547IslandingCategory' ,
    'IEEE1547NormalPerformanceCategory' ,
    'IEEE1547Setting' ,
    'IEEE1547TripSettings' ,
    'IOPoint' ,
    'IdentifiedObject' ,
    'IncrementalHeatRateCurve' ,
    'InflowForecast' ,
    'InspectionDataSet' ,
    'IntegerQuantity' ,
    'InterrupterUnit' ,
    'InterrupterUnitInfo' ,
    'InterruptingMediumKind' ,
    'IntervalBlock' ,
    'IntervalReading' ,
    'IrregularIntervalSchedule' ,
    'IrregularTimePoint' ,
    'Issuer' ,
    'Jumper' ,
    'Junction' ,
    'LevelVsVolumeCurve' ,
    'Limit' ,
    'LimitSet' ,
    'Line' ,
    'LineFault' ,
    'LinearShuntCompensator' ,
    'LinearShuntCompensatorPhase' ,
    'LoadArea' ,
    'LoadBreakSwitch' ,
    'LoadGroup' ,
    'LoadResponseCharacteristic' ,
    'Location' ,
    'LocationKind' ,
    'MacroPeriodKind' ,
    'Manufacturer' ,
    'Measurement' ,
    'MeasurementKind' ,
    'MeasurementValue' ,
    'MeasurementValueQuality' ,
    'MeasurementValueSource' ,
    'MeasuringPeriodKind' ,
    'Medium' ,
    'MediumKind' ,
    'Meter' ,
    'MeterMultiplier' ,
    'MeterMultiplierKind' ,
    'MeterReading' ,
    'MeterWorkTask' ,
    'MetrologyRequirement' ,
    'MonthDayInterval' ,
    'MutualCoupling' ,
    'Name' ,
    'NameType' ,
    'NameTypeAuthority' ,
    'NamedPhenomenon' ,
    'NoLoadTest' ,
    'NonConformLoad' ,
    'NonConformLoadGroup' ,
    'NonConformLoadSchedule' ,
    'NonlinearShuntCompensator' ,
    'NonlinearShuntCompensatorPhase' ,
    'NonlinearShuntCompensatorPhasePoint' ,
    'NonlinearShuntCompensatorPoint' ,
    'NotificationTriggerKind' ,
    'NuclearGeneratingUnit' ,
    'OpenCircuitTest' ,
    'OperatingMechanism' ,
    'OperatingMechanismInfo' ,
    'OperatingMechanismKind' ,
    'OperatingParticipant' ,
    'OperatingShare' ,
    'OperationalLimit' ,
    'OperationalLimitDirectionKind' ,
    'OperationalLimitSet' ,
    'OperationalLimitType' ,
    'Organisation' ,
    'OrganisationRole' ,
    'OverfrequencyTripCurve' ,
    'OverheadWireInfo' ,
    'OvervoltageTripCurve' ,
    'Ownership' ,
    'PSRType' ,
    'ParallelLineSegment' ,
    'ParentOrganization' ,
    'PendingCalculation' ,
    'PenstockLossCurve' ,
    'PerLengthImpedance' ,
    'PerLengthLineParameter' ,
    'PerLengthPhaseImpedance' ,
    'PerLengthSequenceImpedance' ,
    'Person' ,
    'PersonRole' ,
    'PetersenCoil' ,
    'PetersenCoilModeKind' ,
    'PhaseCode' ,
    'PhaseConnectedFaultKind' ,
    'PhaseImpedanceData' ,
    'PhaseShuntConnectionKind' ,
    'PhaseTapChanger' ,
    'PhaseTapChangerAsymmetrical' ,
    'PhaseTapChangerLinear' ,
    'PhaseTapChangerNonLinear' ,
    'PhaseTapChangerSymmetrical' ,
    'PhaseTapChangerTable' ,
    'PhaseTapChangerTablePoint' ,
    'PhaseTapChangerTabular' ,
    'PhenomenonClassification' ,
    'PhotovoltaicUnit' ,
    'Plant' ,
    'PositionPoint' ,
    'PowerCutZone' ,
    'PowerElectronicsConnection' ,
    'PowerElectronicsConnectionPhase' ,
    'PowerElectronicsUnit' ,
    'PowerElectronicsWindUnit' ,
    'PowerSystemResource' ,
    'PowerTransformer' ,
    'PowerTransformerEnd' ,
    'PowerTransformerInfo' ,
    'PricingStructure' ,
    'Priority' ,
    'Procedure' ,
    'ProcedureDataSet' ,
    'ProcedureKind' ,
    'ProductAssetModel' ,
    'ProtectedSwitch' ,
    'Quality61850' ,
    'RaiseLowerCommand' ,
    'RandomisationKind' ,
    'RatioTapChanger' ,
    'RatioTapChangerTable' ,
    'RatioTapChangerTablePoint' ,
    'RationalNumber' ,
    'ReactiveCapabilityCurve' ,
    'Reading' ,
    'ReadingInterharmonic' ,
    'ReadingReasonKind' ,
    'ReadingType' ,
    'Recloser' ,
    'Register' ,
    'RegularIntervalSchedule' ,
    'RegularTimePoint' ,
    'RegulatingCondEq' ,
    'RegulatingControl' ,
    'RegulatingControlModeKind' ,
    'RegulationSchedule' ,
    'ReportingCapability' ,
    'ReportingGroup' ,
    'ReportingMethodKind' ,
    'ReportingSuperGroup' ,
    'Reservoir' ,
    'RightOfWay' ,
    'RotatingMachine' ,
    'SVCControlMode' ,
    'ScaleKind' ,
    'ScheduledEvent' ,
    'ScheduledEventData' ,
    'Seal' ,
    'SealConditionKind' ,
    'SealKind' ,
    'Season' ,
    'SeasonDayTypeSchedule' ,
    'Sectionaliser' ,
    'SeriesCompensator' ,
    'ServiceCategory' ,
    'ServiceKind' ,
    'ServiceLocation' ,
    'ServiceMultiplier' ,
    'ServiceMultiplierKind' ,
    'SetPoint' ,
    'ShortCircuitRotorKind' ,
    'ShortCircuitTest' ,
    'ShuntCompensator' ,
    'ShuntCompensatorInfo' ,
    'ShuntCompensatorPhase' ,
    'ShutdownCurve' ,
    'SinglePhaseKind' ,
    'SmartInverterMode' ,
    'SolarGeneratingUnit' ,
    'StartIgnFuelCurve' ,
    'StartMainFuelCurve' ,
    'StartRampCurve' ,
    'StartupModel' ,
    'StateVariable' ,
    'StaticVarCompensator' ,
    'StationSupply' ,
    'Status' ,
    'SteamSendoutSchedule' ,
    'StreetAddress' ,
    'StreetDetail' ,
    'StringMeasurement' ,
    'StringMeasurementValue' ,
    'StringQuantity' ,
    'Structure' ,
    'StructureSupport' ,
    'SubGeographicalRegion' ,
    'SubLoadArea' ,
    'Substation' ,
    'SvEstVoltage' ,
    'SvInjection' ,
    'SvPowerFlow' ,
    'SvShuntCompensatorSections' ,
    'SvStatus' ,
    'SvSwitch' ,
    'SvTapStep' ,
    'SvVoltage' ,
    'Switch' ,
    'SwitchInfo' ,
    'SwitchOperationSummary' ,
    'SwitchPhase' ,
    'SwitchSchedule' ,
    'SynchronousMachine' ,
    'SynchronousMachineKind' ,
    'SynchronousMachineOperatingMode' ,
    'TailbayLossCurve' ,
    'TapChanger' ,
    'TapChangerControl' ,
    'TapChangerInfo' ,
    'TapChangerTablePoint' ,
    'TapSchedule' ,
    'TapeShieldCableInfo' ,
    'TargetLevelSchedule' ,
    'TelephoneNumber' ,
    'Terminal' ,
    'TestKind' ,
    'ThermalGeneratingUnit' ,
    'ThermostatControlMode' ,
    'ThermostatController' ,
    'TimeInterval' ,
    'TimeIntervalKind' ,
    'TopologicalIsland' ,
    'TopologicalNode' ,
    'TownDetail' ,
    'TransformerApplicationKind' ,
    'TransformerControlMode' ,
    'TransformerCoreAdmittance' ,
    'TransformerEnd' ,
    'TransformerEndInfo' ,
    'TransformerMeshImpedance' ,
    'TransformerStarImpedance' ,
    'TransformerTank' ,
    'TransformerTankEnd' ,
    'TransformerTankInfo' ,
    'TransformerTest' ,
    'TroubleTicket' ,
    'UnderfrequencyTripCurve' ,
    'UndervoltageTripCurve' ,
    'UnitMultiplier' ,
    'UnitSymbol' ,
    'UsagePoint' ,
    'UsagePointConnectedKind' ,
    'UsagePointGroup' ,
    'UsagePointLocation' ,
    'UserAttribute' ,
    'Validity' ,
    'ValueAliasSet' ,
    'ValueToAlias' ,
    'Version' ,
    'VoltVarCurve' ,
    'VoltWattCurve' ,
    'VoltageControlZone' ,
    'VoltageLevel' ,
    'VoltageLimit' ,
    'WattVarCurve' ,
    'WindGenUnitKind' ,
    'WindGeneratingUnit' ,
    'WindingConnection' ,
    'WireAssemblyInfo' ,
    'WireInfo' ,
    'WireInsulationKind' ,
    'WireMaterialKind' ,
    'WirePhaseInfo' ,
    'WirePosition' ,
    'WireSpacingInfo' ,
    'WireUsageKind' ,
    'MonthDay' ,
    'Susceptance' ,
    'Pressure' ,
    'Hours' ,
    'Capacitance' ,
    'Voltage' ,
    'SusceptancePerLength' ,
    'Frequency' ,
    'PU' ,
    'CalculationIntervalUnitKind' ,
    'Minutes' ,
    'AngleDegrees' ,
    'ActivePowerChangeRate' ,
    'RotationSpeed' ,
    'RealEnergy' ,
    'VolumeFlowRate' ,
    'StructureMaterialKind' ,
    'CostPerHeatUnit' ,
    'Money' ,
    'Area' ,
    'AnchorKind' ,
    'Volume' ,
    'CostPerEnergyUnit' ,
    'KiloActivePower' ,
    'Displacement' ,
    'ApparentPower' ,
    'CurrentFlow' ,
    'Source' ,
    'ReactivePower' ,
    'Temperature' ,
    'Seconds' ,
    'ActivePowerPerFrequency' ,
    'Emission' ,
    'VoltagePerReactivePower' ,
    'WaterLevel' ,
    'ActivePower' ,
    'Mass' ,
    'ConductancePerLength' ,
    'Reactance' ,
    'Conductance' ,
    'ReactancePerLength' ,
    'Resistance' ,
    'CostPerVolume' ,
    'StructureSupportKind' ,
    'CostRate' ,
    'Length' ,
    'ResistancePerLength' ,
    'PerCent' ,
    'AngleRadians' ,
    'HeatRate' ,
    'Impedance'
]



from __future__ import annotations

import importlib
import json
import logging
import math
import os
from uuid import UUID

from rdflib import Graph, Namespace, URIRef
from rdflib.namespace import RDF

import cimgraph.queries.sparql as sparql
from cimgraph.databases import (ConnectionInterface, ConnectionParameters,
                                QueryResponse)
from cimgraph.models.graph_model import GraphModel

_log = logging.getLogger(__name__)


class RDFlibConnection(ConnectionInterface):

    def __init__(self,
                 connection_params: ConnectionParameters,
                 use_oxigraph: bool = True):
        self.cim_profile = connection_params.cim_profile
        self.cim = importlib.import_module('cimgraph.data_profile.' + self.cim_profile)
        self.namespace = connection_params.namespace
        self.iec61970_301 = connection_params.iec61970_301
        self.filename = connection_params.filename
        self.connection_params = connection_params
        self.libgraph = None
        self.use_oxigraph = use_oxigraph

    def connect(self):
        if not self.libgraph:
            if self.use_oxigraph:
                self.libgraph = Graph(store='Oxigraph')
            else:
                self.libgraph = Graph()

            if self.filename is not None:
                try:
                    self.libgraph.parse(self.filename)
                    self.libgraph.bind('cim', Namespace(self.namespace))
                    self.libgraph.bind('rdf', RDF)
                except:
                    _log.warning(f'File {self.filename} not found. Defaulting to empty network graph')
                    self.filename = None

    def disconnect(self):
        self.libgraph = None

    def execute(self, query_message: str) -> QueryResponse:
        self.connect()
        query_output = self.libgraph.query(query_message)
        return query_output

    def get_object(self, mrid:str, graph = {}) -> object:
        sparql_message = sparql.get_object_sparql(mrid, self.connection_params)
        query_output = self.execute(sparql_message)
        obj = None
        for result in query_output:
            uri = result['identifier']['value']
            obj_class = result['obj_class']['value']
            class_type = eval(f'self.cim.{obj_class}')
            obj = self.create_object(graph, class_type, uri)
        return obj

    def create_new_graph(self, container: object) -> dict[type, dict[UUID, object]]:
        graph = {}
        self.add_to_graph(graph=graph, obj=container)
        # Get all nodes, terminal, and equipment by
        sparql_message = sparql.get_all_nodes_from_container(container, self.connection_params)
        query_output = self.execute(sparql_message)
        graph = self.parse_node_query(graph, query_output)
        return graph

    def parse_node_query(self, graph: dict, query_output: dict) -> dict[type, dict[UUID, object]]:

        for result in query_output:
            # Parse query results
            node_mrid = str(result.ConnectivityNode)
            term_mrid = str(result.Terminal)
            eq = json.loads(result.Equipment)
            eq_id = eq['@id']
            eq_class = eq['@type']

            if eq_class in self.cim.__all__:
                eq_class = eval(f'self.cim.{eq_class}')
                equipment = self.create_object(graph, eq_class, eq_id)

            else:
                _log.warning(f'object class missing from data profile: {eq_class}')
                continue

            if node_mrid != 'None':
                # Add each object to graph
                node = self.create_object(graph, self.cim.ConnectivityNode, node_mrid)
                terminal = self.create_object(graph, self.cim.Terminal, term_mrid)
                # Associate the node and equipment with the terminal
                if terminal not in equipment.Terminals:
                    equipment.Terminals.append(terminal)
                if terminal not in node.Terminals:
                    node.Terminals.append(terminal)
                # Associate the terminal with the equipment and node
                setattr(terminal, 'ConnectivityNode', node)
                setattr(terminal, 'ConductingEquipment', equipment)

        return graph

    def get_edges_query(self, graph: dict[type, dict[UUID, object]],
                        cim_class: type) -> str:

        eq_mrids = list(graph[cim_class].keys())[0:100]
        sparql_message = sparql.get_all_edges_sparql(graph, cim_class, eq_mrids,
                                                     self.connection_params)

        return sparql_message

    def get_all_edges(self, graph: dict[type, dict[UUID, object]], cim_class: type) -> None:
        uuid_list = list(graph[cim_class].keys())
        for index in range(math.ceil(len(uuid_list) / 100)):
            eq_mrids = uuid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct queries>sparql python script based on class name
            sparql_message = sparql.get_all_edges_sparql(graph,
                cim_class, eq_mrids, self.connection_params)
            #execute sparql query
            query_output = self.execute(sparql_message)
            self.edge_query_parser(query_output, graph, cim_class)

    def get_all_attributes(self, graph: dict[type, dict[UUID, object]],
                           cim_class: type) -> None:
        mrid_list = list(graph[cim_class].keys())

        for index in range(math.ceil(len(mrid_list) / 100)):
            eq_mrids = mrid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct loaders>sparql python script based on class name
            sparql_message = sparql.get_all_attributes_sparql(graph,
                cim_class, eq_mrids, self.connection_params)
            #execute sparql query
            query_output = self.execute(sparql_message)
            self.edge_query_parser(query_output, graph, cim_class, expand_graph = False)

    def edge_query_parser(self, query_output: QueryResponse,
                          graph: dict[type, dict[UUID, object]],
                          cim_class: type, expand_graph = True) -> None:
        for result in query_output:
            if 'type' not in result.attribute and result.attribute is not None:  #skip 'type' and other single attributes

                is_association = False
                is_enumeration = False
                if result.mRID is not None:  #get mRID
                    mRID = str(result.mRID)
                else:
                    iri = str(result.eq)
                    if self.iec61970_301 > 7:
                        mRID = iri.split('uuid:')[1]
                    else:
                        mRID = iri.split('rdf:id:')[1]
                identifier = UUID(mRID.strip('_').lower())
                attr_uri = result.attr
                attr = str(result.attr).split(self.namespace)[1]
                attribute = attr.split('.')  #split edge attribute
                value = str(result.val)  #get edge value

                if self.namespace in value:  #check if enumeration
                    enum_text = value.split(self.namespace)[1]
                    enum_text = enum_text.split('>')[0]
                    enum_class = enum_text.split('.')[0]
                    enum_value = enum_text.split('.')[1]
                    is_enumeration = True

                if result.edge_class is not None:  #check if association
                    is_association = True
                    # edge = json.loads(result.edge)
                    # edge_mRID = edge['@id']
                    # edge_class = edge['@type']
                    edge_class = str(result.edge_class)
                    if result.edge_mRID is not None:
                        edge_mRID = str(result.edge_mRID)
                    else:
                        if self.iec61970_301 > 7:
                            edge_mRID = value.split('uuid:')[1]
                        else:
                            edge_mRID = value.split('#')[1]
                    if edge_class in self.cim.__all__:
                        edge_class = eval(f'self.cim.{edge_class}')
                    else:
                        _log.warning(f'unknown class {edge_class}')
                        continue

                    if expand_graph:
                        self.create_edge(graph, cim_class, identifier, attribute, edge_class, edge_mRID)
                    else:
                        self.create_value(graph, cim_class, identifier, attribute, value)


                elif is_enumeration:
                    edge_enum = eval(f'self.cim.{enum_class}(enum_value)')
                    association = self.check_attribute(cim_class, attribute)
                    if association is not None:
                        setattr(graph[cim_class][identifier], association, edge_enum)
                else:
                    association = self.check_attribute(cim_class, attribute)
                    if association is not None:
                        self.create_value(graph, cim_class, identifier, attribute, value)


    #     return obj

    def upload(self, graph):
        pass


    def get_from_triple(self, subject, predicate, graph = ...):
        pass

    def create_distributed_graph(self, area, graph = ...):
        pass
